from deciphon_schema import SeqRead
from fastapi import APIRouter, Request
from starlette.status import HTTP_200_OK

from deciphon_sched.database import Database
from deciphon_sched.errors import NotFoundInDatabaseError
from deciphon_sched.sched.models import Seq

router = APIRouter()


@router.get("/seqs", status_code=HTTP_200_OK)
async def read_seqs(request: Request) -> list[SeqRead]:
    database: Database = request.app.state.database
    with database.create_session() as session:
        return [x.read_model() for x in Seq.get_all(session)]


@router.get("/seqs/{seq_id}", status_code=HTTP_200_OK)
async def read_seq(request: Request, seq_id: int) -> SeqRead:
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Seq.get_by_id(session, seq_id)
        if x is None:
            raise NotFoundInDatabaseError("Seq")
        return x.read_model()



import pytest

from deciphon_sched.journal import Journal
from deciphon_sched.logger import Logger
from deciphon_sched.settings import Settings


@pytest.mark.asyncio
async def test_journal(mqtt, settings: Settings, logger: Logger):
    settings.mqtt_host = mqtt["host"]
    settings.mqtt_port = mqtt["port"]
    async with Journal(settings, logger) as journal:
        await journal.publish("press", "1")



from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime, timezone
from typing import Optional

from deciphon_schema import (
    DBName,
    DBRead,
    HMMName,
    HMMRead,
    JobRead,
    JobState,
    JobType,
    ScanRead,
    SeqRead,
    SnapRead,
)
from sqlalchemy import ForeignKey, select
from sqlalchemy.orm import DeclarativeBase, Mapped, Session, mapped_column, relationship

from deciphon_sched.errors import JobStateTransitionError


def metadata():
    return BaseModel.metadata


class BaseModel(DeclarativeBase):
    def model_dump(self):
        return {field.name: getattr(self, field.name) for field in self.__table__.c}

    def __repr__(self):
        params = ", ".join(f"{k}={v}" for k, v in self.model_dump().items())
        return f"{self.__class__.__name__}({params})"

    def __str__(self):
        return repr(self)


DELETE = "save-update, merge, delete"


class Job(BaseModel):
    __tablename__ = "job"

    id: Mapped[int] = mapped_column(primary_key=True)

    hmm: Mapped[Optional[HMM]] = relationship(back_populates="job")
    scan: Mapped[Optional[Scan]] = relationship(back_populates="job")

    type: Mapped[JobType]
    state: Mapped[JobState]
    progress: Mapped[int]
    error: Mapped[str]
    submission: Mapped[datetime]
    exec_started: Mapped[Optional[datetime]]
    exec_ended: Mapped[Optional[datetime]]

    @classmethod
    def create(cls, type: JobType):
        return cls(
            type=type,
            state=JobState.pend,
            progress=0,
            error="",
            submission=datetime.now(timezone.utc),
        )

    def set_pend(self):
        raise JobStateTransitionError(str(self.state), str(JobState.pend))

    def set_run(self, progress: int):
        if self.state in [JobState.done, JobState.fail]:
            raise JobStateTransitionError(str(self.state), str(JobState.run))
        self.state = JobState.run
        self.progress = progress
        if self.exec_started is None:
            self.exec_started = datetime.now()

    def set_done(self):
        if self.state in [JobState.done, JobState.fail]:
            raise JobStateTransitionError(str(self.state), str(JobState.done))
        self.state = JobState.done
        self.progress = 100
        if self.exec_started is None:
            self.exec_started = datetime.now()
        self.exec_ended = datetime.now()

    def set_fail(self, error: str):
        if self.state in [JobState.done, JobState.fail]:
            raise JobStateTransitionError(str(self.state), str(JobState.fail))
        self.state = JobState.fail
        self.error = error
        if self.exec_started is None:
            self.exec_started = datetime.now()
        self.exec_ended = datetime.now()

    def read_model(self):
        return JobRead(
            id=self.id,
            type=self.type,
            state=self.state,
            progress=self.progress,
            error=self.error,
            submission=self.submission,
            exec_started=self.exec_started,
            exec_ended=self.exec_ended,
        )

    @staticmethod
    def get_by_id(session: Session, id: int):
        x = session.execute(select(Job).where(Job.id == id)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_all(session: Session):
        return [x._tuple()[0] for x in session.execute(select(Job)).all()]


class HMM(BaseModel):
    __tablename__ = "hmm"

    id: Mapped[int] = mapped_column(primary_key=True)
    job_id: Mapped[int] = mapped_column(ForeignKey("job.id"))

    job: Mapped[Job] = relationship(back_populates="hmm", cascade=DELETE)
    db: Mapped[Optional[DB]] = relationship(back_populates="hmm")

    name: Mapped[str] = mapped_column(unique=True)

    @classmethod
    def create(cls, file: HMMName):
        return cls(job=Job.create(type=JobType.hmm), name=file.name)

    def read_model(self):
        file = HMMName(name=self.name)
        return HMMRead(id=self.id, job=self.job.read_model(), file=file)

    @staticmethod
    def get_by_id(session: Session, id: int):
        x = session.execute(select(HMM).where(HMM.id == id)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_by_name(session: Session, hmmname: HMMName):
        x = session.execute(select(HMM).where(HMM.name == hmmname.name)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_all(session: Session):
        return [x._tuple()[0] for x in session.execute(select(HMM)).all()]


class DB(BaseModel):
    __tablename__ = "db"

    id: Mapped[int] = mapped_column(primary_key=True)
    hmm_id: Mapped[int] = mapped_column(ForeignKey("hmm.id"))

    hmm: Mapped[HMM] = relationship(back_populates="db")
    scans: Mapped[list[Scan]] = relationship(back_populates="db")

    name: Mapped[str]

    @classmethod
    def create(cls, hmm: HMM, file: DBName):
        return cls(hmm=hmm, name=file.name)

    def read_model(self):
        file = DBName(name=self.name)
        return DBRead(id=self.id, hmm=self.hmm.read_model(), file=file)

    @staticmethod
    def get_by_id(session: Session, id: int):
        x = session.execute(select(DB).where(DB.id == id)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_by_name(session: Session, hmmname: DBName):
        x = session.execute(select(DB).where(DB.name == hmmname.name)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_all(session: Session):
        return [x._tuple()[0] for x in session.execute(select(DB)).all()]


class Seq(BaseModel):
    __tablename__ = "seq"

    id: Mapped[int] = mapped_column(primary_key=True)
    scan_id: Mapped[int] = mapped_column(ForeignKey("scan.id"))

    scan: Mapped[Scan] = relationship(back_populates="seqs")

    name: Mapped[str]
    data: Mapped[str]

    @classmethod
    def create(cls, name: str, data: str):
        return cls(name=name, data=data)

    def read_model(self):
        return SeqRead(id=self.id, name=self.name, data=self.data)

    @staticmethod
    def get_by_id(session: Session, id: int):
        x = session.execute(select(Seq).where(Seq.id == id)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_all(session: Session):
        return [x._tuple()[0] for x in session.execute(select(Seq)).all()]


class Snap(BaseModel):
    __tablename__ = "snap"

    id: Mapped[int] = mapped_column(primary_key=True)
    scan_id: Mapped[int] = mapped_column(ForeignKey("scan.id"))

    scan: Mapped[Scan] = relationship(back_populates="snap")

    data: Mapped[bytes]

    @classmethod
    def create(cls, scan_id: int, data: bytes):
        return cls(scan_id=scan_id, data=data)

    def read_model(self):
        return SnapRead(id=self.id, size=len(self.data))

    @staticmethod
    def get_by_scan_id(session: Session, id: int):
        x = session.execute(select(Snap).where(Snap.scan_id == id)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_all(session: Session):
        return [x._tuple()[0] for x in session.execute(select(Snap)).all()]


class Scan(BaseModel):
    __tablename__ = "scan"

    id: Mapped[int] = mapped_column(primary_key=True)
    job_id: Mapped[int] = mapped_column(ForeignKey("job.id"))
    db_id: Mapped[int] = mapped_column(ForeignKey("db.id"))

    multi_hits: Mapped[bool]
    hmmer3_compat: Mapped[bool]

    job: Mapped[Job] = relationship(back_populates="scan", cascade=DELETE)
    seqs: Mapped[list[Seq]] = relationship(back_populates="scan", cascade=DELETE)
    snap: Mapped[Optional[Snap]] = relationship(back_populates="scan", cascade=DELETE)
    db: Mapped[DB] = relationship(back_populates="scans")

    @classmethod
    def create(cls, db: DB, multi_hits: bool, hmmer3_compat: bool, seqs: Iterable[Seq]):
        return cls(
            db=db,
            multi_hits=multi_hits,
            hmmer3_compat=hmmer3_compat,
            job=Job.create(type=JobType.scan),
            seqs=list(seqs),
        )

    def read_model(self):
        return ScanRead(
            id=self.id,
            job=self.job.read_model(),
            db=self.db.read_model(),
            multi_hits=self.multi_hits,
            hmmer3_compat=self.hmmer3_compat,
            seqs=[x.read_model() for x in self.seqs],
        )

    @staticmethod
    def get_by_id(session: Session, id: int):
        x = session.execute(select(Scan).where(Scan.id == id)).one_or_none()
        return x if x is None else x._tuple()[0]

    @staticmethod
    def get_all(session: Session, job_id: Optional[int] = None):
        clause = select(Scan)
        if job_id is not None:
            clause = clause.where(Scan.job_id == job_id)
        return [x._tuple()[0] for x in session.execute(clause).all()]



import os
from functools import partial
from threading import Thread

import uvicorn
from typer import Option, Typer
from typing_extensions import Annotated

from deciphon_sched.main import create_app
from deciphon_sched.settings import Settings
from deciphon_sched.signals import raise_sigint_on_sigterm, sigint_hook

app = Typer()

RELOAD = Annotated[bool, Option(help="Enable auto-reload.")]


def wrap(server: uvicorn.Server):
    server.run()


@app.command()
def main(reload: RELOAD = False):
    raise_sigint_on_sigterm()

    settings = Settings()
    config = uvicorn.Config(
        partial(create_app, settings),
        host=settings.host,
        port=settings.port,
        log_level=settings.log_level.value,
        reload=reload,
        factory=True,
    )
    server = uvicorn.Server(config)

    # Send uvicorn to a subthread to prevent it from capturing signals.
    thread = Thread(target=partial(wrap, server))

    def shutdown(server: uvicorn.Server):
        if server.should_exit:
            os.abort()
        server.should_exit = True

    sigint_hook(partial(shutdown, server))
    thread.start()
    thread.join()



from pathlib import Path

import requests
from pydantic import HttpUrl

from deciphon_sched.logger import Logger
from deciphon_sched.settings import Settings
from deciphon_sched.storage import Storage


def test_storage(s3, settings: Settings, logger: Logger, tmp_path: Path):
    settings.s3_key = s3["access_key"]
    settings.s3_secret = s3["secret_key"]
    settings.s3_url = HttpUrl(s3["url"])
    storage = Storage(settings, logger)

    info = storage.presigned_upload("example.txt")

    with open(tmp_path / "example.txt", "w") as f:
        f.write("content")

    with open(tmp_path / "example.txt", "rb") as f:
        files = {"file": (str(tmp_path / "example.txt"), f)}
        url = info.url.unicode_string()
        response = requests.post(url, data=info.fields, files=files)
        assert response.ok
        assert response.status_code == 204



import pytest

from deciphon_sched.logger import Logger
from deciphon_sched.settings import Settings
from deciphon_sched.testing import mqtt_server, s3_cleanup, s3_server


@pytest.fixture
def settings():
    return Settings()


@pytest.fixture
def logger(settings: Settings):
    return Logger(settings)


mqtt = pytest.fixture(mqtt_server, scope="package")
s3_server = pytest.fixture(s3_server, scope="package")


@pytest.fixture
def s3(s3_server):
    s3_cleanup(s3_server["client"])
    yield s3_server



import requests
from fastapi.testclient import TestClient


def test_hmm_not_found(client: TestClient):
    assert client.get("/hmms/1").status_code == 404


def test_presigned_upload(client: TestClient):
    assert client.get("/hmms/presigned-upload/minifam.hmm").status_code == 200


def test_presigned_download_failure(client: TestClient):
    assert client.get("/hmms/presigned-download/minifam.hmm").status_code == 422


def test_create(client: TestClient, files, s3_upload):
    response = client.get("/hmms/presigned-upload/minifam.hmm")
    s3_upload(response.json(), files / "minifam.hmm")
    assert upload_hmm(client, s3_upload, files / "minifam.hmm").status_code == 201


def test_download(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    url = client.get("/hmms/presigned-download/minifam.hmm").json()["url"]
    response = requests.get(url)
    assert response.status_code == 200
    assert hash(open(files / "minifam.hmm", "rb").read()) == hash(response.text)


def test_read_one(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    assert client.get("/hmms/1").status_code == 200


def test_read_many(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    assert client.get("/hmms").status_code == 200


def test_delete(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    assert client.delete("/hmms/1").status_code == 204


def upload_hmm(client: TestClient, s3_upload, file):
    response = client.get(f"/hmms/presigned-upload/{file.name}")
    s3_upload(response.json(), file)
    params = {"gencode": 1, "epsilon": 0.01}
    return client.post("/hmms/", params=params, json={"name": file.name})



import pathlib

import pytest
import requests
from fastapi.testclient import TestClient

from deciphon_sched.main import create_app
from deciphon_sched.settings import Settings


@pytest.fixture
def s3_upload():
    def upload(presigned_upload, file):
        with open(file, "rb") as f:
            files = {"file": (file.name, f)}
            url = presigned_upload["url"]
            fields = presigned_upload["fields"]
            http_response = requests.post(url, data=fields, files=files)
            assert http_response.status_code == 204

    yield upload


@pytest.fixture
def compose(mqtt, s3, settings: Settings):
    data = settings.model_dump()
    data["mqtt_host"] = mqtt["host"]
    data["mqtt_port"] = mqtt["port"]
    data["s3_key"] = s3["access_key"]
    data["s3_secret"] = s3["secret_key"]
    data["s3_url"] = s3["url"]
    settings = Settings.model_validate(data)
    yield create_app(settings)


@pytest.fixture
def client(compose):
    with TestClient(compose) as client:
        yield client


@pytest.fixture()
def files() -> pathlib.Path:
    return pathlib.Path(__file__).parent / "files"



import os
import signal
from typing import Callable


def ignore_sigint():
    # https://stackoverflow.com/q/1408356
    # We avoid the KeyboardInterrupt bug by
    # ignore SIGINT altogether at the child
    # process
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def raise_sigint(*_):
    os.kill(os.getpid(), signal.SIGINT)


def raise_sigint_on_sigterm(*_):
    signal.signal(signal.SIGTERM, raise_sigint)


def sigint_hook(callback: Callable[[], None]):
    signal.signal(signal.SIGINT, lambda *_: callback())
    signal.signal(signal.SIGTERM, lambda *_: callback())



from deciphon_sched.database import Database
from deciphon_sched.logger import Logger
from deciphon_sched.sched.models import metadata
from deciphon_sched.settings import Settings

TABLE_NAMES = ["job", "hmm", "db", "scan", "seq", "snap"]


def test_database_tables(settings: Settings, logger: Logger):
    database = Database(settings, logger)
    database.create_tables(metadata())
    table_names = [x for x in database.metadata().tables]
    assert set(table_names) == set(TABLE_NAMES)



import pytest
from deciphon_schema import DBName, HMMName
from sqlalchemy import select

from deciphon_sched.database import Database
from deciphon_sched.logger import Logger
from deciphon_sched.sched.models import (
    DB,
    HMM,
    Scan,
    Seq,
    metadata,
)
from deciphon_sched.settings import Settings


@pytest.fixture()
def hmmfile():
    return HMMName(name="file.hmm")


@pytest.fixture()
def dbfile():
    return DBName(name="file.dcp")


@pytest.fixture()
def session(settings: Settings, logger: Logger):
    database = Database(settings, logger)
    database.create_tables(metadata())
    yield database.create_session()


def test_models_add_get_hmm(session, hmmfile: HMMName):
    for i in range(1, 2):
        hmm = HMM.create(file=hmmfile)

        session.add(hmm)
        session.commit()
        assert hmm.id == i
        assert hmm.job.id == i

        hmm = session.scalars(select(HMM).where(HMM.id == hmm.id)).one()
        assert hmm.id == i
        assert hmm.job.id == i


def test_models_add_get_db(session, hmmfile: HMMName, dbfile: DBName):
    for i in range(1, 2):
        db = DB.create(hmm=HMM.create(file=hmmfile), file=dbfile)

        session.add(db)
        session.commit()
        assert db.id == i
        assert db.hmm.id == i
        assert db.hmm.job.id == i

        db = session.scalars(select(DB).where(DB.id == db.id)).one()
        assert db.id == i
        assert db.hmm.id == i
        assert db.hmm.job.id == i


def test_models_add_get_scan(session, hmmfile: HMMName, dbfile: DBName):
    for i in range(1, 2):
        db = DB.create(hmm=HMM.create(file=hmmfile), file=dbfile)
        seqs = [Seq(name="seq1", data="ACGT"), Seq(name="seq2", data="CGA")]
        scan = Scan.create(db, True, False, seqs)

        session.add(scan)
        session.commit()
        assert scan.id == i
        assert scan.job.id == i
        assert scan.db.id == i
        assert scan.db.hmm.id == i
        assert scan.db.hmm.job.id == i * 2
        assert scan.seqs[0].scan.id == i
        assert scan.seqs[1].scan.id == i
        assert scan.seqs[0].name == seqs[0].name
        assert scan.seqs[1].name == seqs[1].name
        assert seqs[0].id == 1
        assert seqs[1].id == 2
        assert seqs[0].scan_id == scan.id
        assert seqs[1].scan_id == scan.id



from typing import Annotated

from deciphon_schema import DBName, DBRead
from fastapi import APIRouter, Request
from starlette.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT

from deciphon_sched.database import Database
from deciphon_sched.errors import (
    FileNameExistsError,
    FileNameNotFoundError,
    NotFoundInDatabaseError,
)
from deciphon_sched.sched.models import DB, HMM
from deciphon_sched.sched.schemas import DBFilePathType
from deciphon_sched.storage import PresignedDownload, PresignedUpload, Storage

router = APIRouter()


@router.get("/dbs", status_code=HTTP_200_OK)
async def read_dbs(request: Request) -> list[DBRead]:
    database: Database = request.app.state.database
    with database.create_session() as session:
        return [x.read_model() for x in DB.get_all(session)]


@router.get("/dbs/presigned-upload/{name}", status_code=HTTP_200_OK)
async def presigned_db_upload(
    request: Request,
    name: Annotated[str, DBFilePathType],
) -> PresignedUpload:
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        x = DB.get_by_name(session, DBName(name=name))
        if x is not None:
            raise FileNameExistsError(name)
        return storage.presigned_upload(name)


@router.get("/dbs/presigned-download/{name}", status_code=HTTP_200_OK)
async def presigned_db_download(
    request: Request,
    name: Annotated[str, DBFilePathType],
) -> PresignedDownload:
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        x = DB.get_by_name(session, DBName(name=name))
        if x is None:
            raise FileNameNotFoundError(name)
        return storage.presigned_download(name)


@router.post("/dbs/", status_code=HTTP_201_CREATED)
async def create_db(request: Request, db: DBName) -> DBRead:
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        if DB.get_by_name(session, db) is not None:
            raise FileNameExistsError(db.name)

        hmm = HMM.get_by_name(session, db.hmmname)
        if hmm is None:
            raise FileNameNotFoundError(str(db.hmmname))

        if not storage.has_file(db.name):
            raise FileNameNotFoundError(db.name)

        hmm.job.set_done()

        x = DB.create(hmm, db)
        session.add(x)
        session.commit()
        db_read = x.read_model()

    return db_read


@router.get("/dbs/{db_id}", status_code=HTTP_200_OK)
async def read_db(request: Request, db_id: int) -> DBRead:
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = DB.get_by_id(session, db_id)
        if x is None:
            raise NotFoundInDatabaseError("DB")
        return x.read_model()


@router.delete("/dbs/{db_id}", status_code=HTTP_204_NO_CONTENT)
async def delete_db(request: Request, db_id: int):
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        x = DB.get_by_id(session, db_id)
        if x is None:
            raise NotFoundInDatabaseError("DB")
        if storage.has_file(x.name):
            storage.delete(x.name)
        session.delete(x)
        session.commit()



from functools import partial

from fastapi.testclient import TestClient

SCAN = {
    "db_id": 1,
    "multi_hits": True,
    "hmmer3_compat": True,
    "seqs": [{"name": "seq1", "data": "ACGT"}, {"name": "seq2", "data": "GTT"}],
}


def test_job_not_found(client: TestClient):
    assert client.get("/jobs/1").status_code == 404


def test_read_one(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    assert client.get("/jobs/1").status_code == 200


def test_read_many(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    response = client.get("/jobs")
    assert response.status_code == 200
    assert len(response.json()) == 1


def test_read_limit(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    upload_db(client, s3_upload, files / "minifam.dcp")
    assert client.post("/scans/", json=SCAN).status_code == 201
    response = client.get("/jobs")
    assert response.status_code == 200
    assert len(response.json()) == 2
    response = client.get("/jobs", params={"limit": 3})
    assert response.status_code == 200
    assert len(response.json()) == 2
    response = client.get("/jobs", params={"limit": 2})
    assert response.status_code == 200
    assert len(response.json()) == 2
    response = client.get("/jobs", params={"limit": 1})
    assert response.status_code == 200
    assert len(response.json()) == 1
    response = client.get("/jobs", params={"limit": 0})
    assert response.status_code == 200
    assert len(response.json()) == 0


def test_update_done(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    x = partial(client.patch, "/jobs/1")
    assert x(json={"state": "run", "progress": 0, "error": ""}).status_code == 200
    assert x(json={"state": "run", "progress": 100, "error": ""}).status_code == 200
    assert x(json={"state": "done", "progress": 100, "error": ""}).status_code == 200
    assert x(json={"state": "fail", "progress": 0, "error": ""}).status_code == 400


def test_update_fail(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    x = partial(client.patch, "/jobs/1")
    assert x(json={"state": "run", "progress": 0, "error": ""}).status_code == 200
    assert x(json={"state": "run", "progress": 100, "error": ""}).status_code == 200
    assert x(json={"state": "fail", "progress": 0, "error": "msg"}).status_code == 200
    assert x(json={"state": "done", "progress": 100, "error": ""}).status_code == 400


def upload_hmm(client: TestClient, s3_upload, file):
    response = client.get(f"/hmms/presigned-upload/{file.name}")
    s3_upload(response.json(), file)
    params = {"gencode": 1, "epsilon": 0.01}
    return client.post("/hmms/", params=params, json={"name": file.name})


def upload_db(client: TestClient, s3_upload, file):
    response = client.get(f"/dbs/presigned-upload/{file.name}")
    s3_upload(response.json(), file)
    return client.post("/dbs/", json={"name": file.name})



from fastapi import APIRouter, Request
from starlette.status import HTTP_200_OK

from deciphon_sched.database import Database
from deciphon_sched.journal import Journal
from deciphon_sched.sched.schemas import HealthCheck
from deciphon_sched.storage import Storage

router = APIRouter()


@router.get("/health", status_code=HTTP_200_OK)
async def read_health(request: Request) -> HealthCheck:
    database: Database = request.app.state.database
    database.health_check()

    journal: Journal = request.app.state.journal
    await journal.health_check()

    storage: Storage = request.app.state.storage
    storage.health_check()
    return HealthCheck(status="OK")



from __future__ import annotations

import fastapi
from deciphon_schema import (
    DB_NAME_PATTERN,
    HMM_NAME_PATTERN,
    NAME_MAX_LENGTH,
    DBName,
    JobState,
)
from pydantic import BaseModel


class JobUpdate(BaseModel):
    state: JobState
    progress: int
    error: str


HMMFilePathType = fastapi.Path(
    title="HMM file", pattern=HMM_NAME_PATTERN, max_length=NAME_MAX_LENGTH
)


DBFilePathType = fastapi.Path(
    title="DB file", pattern=DB_NAME_PATTERN, max_length=NAME_MAX_LENGTH
)


class DBCreate(BaseModel):
    file: DBName


class SeqCreate(BaseModel):
    name: str
    data: str


class ScanCreate(BaseModel):
    db_id: int
    multi_hits: bool
    hmmer3_compat: bool
    seqs: list[SeqCreate]


class HealthCheck(BaseModel):
    status: str = "OK"



from contextlib import asynccontextmanager

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from sqlalchemy.exc import IntegrityError

from deciphon_sched.database import Database
from deciphon_sched.errors import integrity_error_handler
from deciphon_sched.journal import Journal
from deciphon_sched.logger import Logger
from deciphon_sched.sched import router
from deciphon_sched.sched.models import metadata
from deciphon_sched.settings import Settings
from deciphon_sched.storage import Storage


@asynccontextmanager
async def lifespan(app: FastAPI):
    settings: Settings = app.state.settings
    logger: Logger = app.state.logger
    app.state.database = Database(settings, logger)
    app.state.database.create_tables(metadata())
    app.state.storage = Storage(settings, logger)
    app.state.journal = Journal(settings, logger)
    async with app.state.journal:
        yield
    app.state.database.dispose()


def create_app(settings: Settings):
    app = FastAPI(lifespan=lifespan, root_path=settings.root_path)
    app.state.settings = settings
    app.state.logger = Logger(settings)

    app.include_router(router, prefix=settings.endpoint_prefix)

    app.add_exception_handler(IntegrityError, integrity_error_handler)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=settings.allow_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    return app



from aiomqtt import Client, MqttError
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_delay,
    wait_exponential,
)

from deciphon_sched.logger import Logger
from deciphon_sched.settings import Settings

TOPIC = "deciphon.org"


async def reconnect(mqtt: Client):
    await mqtt.__aexit__(None, None, None)
    await mqtt.__aenter__()


class Journal:
    def __init__(self, settings: Settings, logger: Logger):
        self._mqtt = Client(hostname=settings.mqtt_host, port=settings.mqtt_port)
        self._logger = logger

    async def __aenter__(self):
        await self._mqtt.__aenter__()
        return self

    async def __aexit__(self, *args, **kargs):
        await self._mqtt.__aexit__(*args, **kargs)

    @retry(
        retry=retry_if_exception_type(MqttError),
        stop=stop_after_delay(10),
        wait=wait_exponential(multiplier=0.001),
    )
    async def publish(self, subject: str, payload: str):
        topic = f"/{TOPIC}/{subject}"
        self._logger.handler.info(f"publishing <{payload}> to <{topic}>")
        try:
            await self._mqtt.publish(topic, payload)
        except MqttError:
            await reconnect(self._mqtt)
            raise

    @retry(
        retry=retry_if_exception_type(MqttError),
        stop=stop_after_delay(10),
        wait=wait_exponential(multiplier=0.001),
    )
    async def health_check(self):
        try:
            await self._mqtt.subscribe("/health")
            await self._mqtt.unsubscribe("/health")
        except MqttError:
            await reconnect(self._mqtt)
            raise



from enum import Enum

from pydantic import AnyUrl, Field, HttpUrl, TypeAdapter
from pydantic_settings import BaseSettings, SettingsConfigDict

from deciphon_sched.url import http_url


class LogLevel(str, Enum):
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"

    # Enum of Python3.10 returns a different string representation.
    # Make it return the same as in Python3.11
    def __str__(self):
        return str(self.value)


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="deciphon_sched_")

    host: str = "0.0.0.0"
    port: int = 8000

    root_path: str = ""
    endpoint_prefix: str = ""
    allow_origins: list[str] = ["http://127.0.0.1", "http://localhost"]
    log_level: LogLevel = Field(default=LogLevel.info)

    database_url: AnyUrl = Field(
        default=TypeAdapter(AnyUrl).validate_strings("sqlite+pysqlite:///:memory:")
    )

    s3_key: str = "minioadmin"
    s3_secret: str = "minioadmin"
    s3_url: HttpUrl = http_url("http://127.0.0.1:9000")

    mqtt_host: str = "127.0.0.1"
    mqtt_port: int = 1883



from fastapi.testclient import TestClient

SCAN = {
    "db_id": 1,
    "multi_hits": True,
    "hmmer3_compat": True,
    "seqs": [{"name": "seq1", "data": "ACGT"}, {"name": "seq2", "data": "GTT"}],
}


def test_create(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    upload_db(client, s3_upload, files / "minifam.dcp")
    assert client.post("/scans/", json=SCAN).status_code == 201


def test_read_one(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    upload_db(client, s3_upload, files / "minifam.dcp")
    client.post("/scans/", json=SCAN)
    assert client.get("/scans/1").status_code == 200


def test_read_many(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    upload_db(client, s3_upload, files / "minifam.dcp")
    client.post("/scans/", json=SCAN)
    assert client.get("/scans").status_code == 200


def test_delete(client: TestClient, files, s3_upload):
    upload_hmm(client, s3_upload, files / "minifam.hmm")
    upload_db(client, s3_upload, files / "minifam.dcp")
    client.post("/scans/", json=SCAN)
    assert client.delete("/scans/1").status_code == 204


def upload_hmm(client: TestClient, s3_upload, file):
    response = client.get(f"/hmms/presigned-upload/{file.name}")
    s3_upload(response.json(), file)
    params = {"gencode": 1, "epsilon": 0.01}
    return client.post("/hmms/", params=params, json={"name": file.name})


def upload_db(client: TestClient, s3_upload, file):
    response = client.get(f"/dbs/presigned-upload/{file.name}")
    s3_upload(response.json(), file)
    return client.post("/dbs/", json={"name": file.name})



import tempfile
from typing import Annotated, Optional

from deciphon_schema import ProdRead, ScanRead, ScanRequest
from deciphon_snap.match import MatchElemName
from deciphon_snap.prod import Prod, ProdList
from deciphon_snap.read_snap import read_snap
from deciphon_snap.view import view_alignments
from fastapi import APIRouter, File, Request, Response
from fastapi.responses import PlainTextResponse
from starlette.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT

from deciphon_sched.database import Database
from deciphon_sched.errors import (
    FoundInDatabaseError,
    NotFoundInDatabaseError,
    SnapFileError,
)
from deciphon_sched.journal import Journal
from deciphon_sched.sched.models import DB, Scan, Seq, Snap
from deciphon_sched.sched.schemas import ScanCreate

router = APIRouter()


@router.get("/scans", status_code=HTTP_200_OK)
async def read_scans(request: Request, job_id: Optional[int] = None) -> list[ScanRead]:
    database: Database = request.app.state.database
    with database.create_session() as session:
        return [x.read_model() for x in Scan.get_all(session, job_id=job_id)]


@router.post("/scans/", status_code=HTTP_201_CREATED)
async def create_scan(request: Request, scan: ScanCreate) -> ScanRead:
    database: Database = request.app.state.database
    with database.create_session() as session:
        db = DB.get_by_id(session, scan.db_id)
        if db is None:
            raise NotFoundInDatabaseError("DB")

        seqs = [Seq.create(x.name, x.data) for x in scan.seqs]
        x = Scan.create(db, scan.multi_hits, scan.hmmer3_compat, seqs)
        for seq in seqs:
            seq.scan = x
        session.add_all([x] + seqs)
        session.flush()
        scan_read = x.read_model()

        journal: Journal = request.app.state.journal
        x = ScanRequest.create(scan_read)

        await journal.publish("scan", x.model_dump_json())
        session.commit()

    return scan_read


@router.get("/scans/{scan_id}", status_code=HTTP_200_OK)
async def read_scan(request: Request, scan_id: int) -> ScanRead:
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Scan.get_by_id(session, scan_id)
        if x is None:
            raise NotFoundInDatabaseError("Scan")
        return x.read_model()


@router.delete("/scans/{scan_id}", status_code=HTTP_204_NO_CONTENT)
async def delete_scan(request: Request, scan_id: int):
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Scan.get_by_id(session, scan_id)
        if x is None:
            raise NotFoundInDatabaseError("Scan")
        session.delete(x)
        session.commit()


@router.post("/scans/{scan_id}/snap.dcs", status_code=HTTP_201_CREATED)
async def upload_snap(request: Request, scan_id: int, file: Annotated[bytes, File()]):
    with tempfile.NamedTemporaryFile() as tmp:
        tmp.write(file)
        tmp.flush()
        try:
            read_snap(tmp.name)
        except Exception as exception:
            raise SnapFileError(str(exception))

    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Snap.get_by_scan_id(session, scan_id)
        if x is not None:
            raise FoundInDatabaseError("Snap")
        snap = Snap.create(scan_id, file)
        session.add(snap)
        session.commit()
        snap.scan.job.set_done()
        session.add(snap)
        session.commit()
        return snap.read_model()


@router.get("/scans/{scan_id}/snap.dcs", status_code=HTTP_200_OK)
async def download_snap(request: Request, scan_id: int):
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Snap.get_by_scan_id(session, scan_id)
        if x is None:
            raise NotFoundInDatabaseError("Snap")
        data = x.data

    headers = {"Content-Disposition": 'attachment; filename="{snap.dcs}"'}
    media_type = "application/octet-stream"
    return Response(data, headers=headers, media_type=media_type)


@router.delete("/scans/{scan_id}/snap.dcs", status_code=HTTP_204_NO_CONTENT)
async def delete_snap(request: Request, scan_id: int):
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Snap.get_by_scan_id(session, scan_id)
        if x is None:
            raise NotFoundInDatabaseError("Snap")
        session.delete(x)
        session.commit()


def get_snap_file(request: Request, scan_id: int):
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = Snap.get_by_scan_id(session, scan_id)
        if x is None:
            raise NotFoundInDatabaseError("Snap")
        data = x.data

        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(data)
            tmp.flush()
            try:
                return read_snap(tmp.name)
            except Exception as exception:
                raise SnapFileError(str(exception))


def make_prod_read(x: Prod):
    return ProdRead(
        seq_id=x.seq_id,
        profile=x.profile,
        abc=x.abc,
        lrt=x.lrt,
        evalue=x.evalue,
    )


@router.get("/scans/{scan_id}/snap.dcs/prods", status_code=HTTP_200_OK)
async def read_prods(request: Request, scan_id: int):
    return [make_prod_read(x) for x in get_snap_file(request, scan_id).products]


def fasta_repr(products: ProdList, name: MatchElemName):
    return PlainTextResponse(products.fasta_list(name).format())


@router.get("/scans/{scan_id}/snap.dcs/queries", status_code=HTTP_200_OK)
async def read_queries(request: Request, scan_id: int):
    return fasta_repr(get_snap_file(request, scan_id).products, MatchElemName.QUERY)


@router.get("/scans/{scan_id}/snap.dcs/states", status_code=HTTP_200_OK)
async def read_states(request: Request, scan_id: int):
    return fasta_repr(get_snap_file(request, scan_id).products, MatchElemName.STATE)


@router.get("/scans/{scan_id}/snap.dcs/codons", status_code=HTTP_200_OK)
async def read_codons(request: Request, scan_id: int):
    return fasta_repr(get_snap_file(request, scan_id).products, MatchElemName.CODON)


@router.get("/scans/{scan_id}/snap.dcs/aminos", status_code=HTTP_200_OK)
async def read_aminos(request: Request, scan_id: int):
    return fasta_repr(get_snap_file(request, scan_id).products, MatchElemName.AMINO)


@router.get("/scans/{scan_id}/snap.dcs/gff", status_code=HTTP_200_OK)
async def read_gff(request: Request, scan_id: int):
    x = get_snap_file(request, scan_id).products.gff_list().format()
    return PlainTextResponse(x)


@router.get("/scans/{scan_id}/snap.dcs/view", status_code=HTTP_200_OK)
async def view_snap(request: Request, scan_id: int):
    x = get_snap_file(request, scan_id)
    txt = "\n".join(view_alignments(x))
    return PlainTextResponse(strip_empty_lines(txt))


def strip_empty_lines(s):
    lines = s.splitlines()
    while lines and not lines[0].strip():
        lines.pop(0)
    while lines and not lines[-1].strip():
        lines.pop()
    return "\n".join(lines) + "\n"
